/**********************************************************************
    duffingPS.c: Stocastic Path Sampling of a Driven Duffing Occilator.

    
    For background on this system, see 
        "Irreversibility of classical fluctuations studied in
            analogue electrical circuits"
    D.G. Luchinsky, P.V.E. McClintock, Nature v389 p463 (1997)
    
    The system under consideration is an overdamped Duffing oscillator driven
    by a periodic force, and white noise.
    
    The equation of motion is 
    
                    dx
                   ---- =K(x,t) + eta(t)
                    dt

                          3
            K(x,t) = x - x  + A cos(w t)
        
            eta(t) is white noise of variance D.
           
    This is a short program to demonstrate that path sampling can be used 
    to study rare fluctuations in this system.

    Positions and the noise are stored every TSTEP for a total time TIME.
    Propagated forward in time:
    
        x[t+1] = x[t] + TSTEP* K(x[t])  + gr[t]+pf[t];
    
    Here K(x) = x-x^3 (the Duffing part of the force), gr[t] is the fluctuating
    force, and pf[t] is the time periodic force.
    
    The system starts in a particular state at t=0, and some condition
    is placed on the system at t=TTEST.  A shooting step involves replacing 
    one of the random forces, and propagating the path forward in time. 
    If the original condition is still met then the path is accepted. 
    
    This program reproduces the system depicted in figure 2a of the 
    above mentioned paper. It requires approximatly 1 minute of computer time
    (on a moderate workstation) per independent sample.
    
    Usage:  duffingPS Number_of_independant_samples
    
    8/98 Gavin Crooks   gavinc@garnet.berkeley.edu
                        http://gold.cchem.berkeley.edu/~gavinc
    
    1/99 Version 2      A few, relativly minor changes.
**********************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdarg.h>


#include "genRan.h"                         /*Random number generator*/

void    die(char *comment, ...);

void propagate(int start, int end);
int initPath();
void relaxPath(int relax);


#define false   (0)
#define true    (1)
#define PI      (3.1415926535898)


#define CYCLES      (3)                 /*Simulate system for CYCLES cycles */          
#define TSTEPSPERC  (8192)              /*Time steps per cycle */               
#define TBINS       (CYCLES*TSTEPSPERC) /*Total number of values that must be stored*/
#define SHOOTS      (TBINS/2)           /*Approximate number of shoots per independent sample*/
#define SKIP        (TSTEPSPERC/128)    /*Ouput the position every SKIP time steps*/

/* System paremeters */
#define D   (0.012)         /*Variance of fluctuating force*/
#define A   (0.264)         /*Strength of the periodic force*/
#define w   (1.2)           /*Time constant of the periodic force*/
int    SAMPLES;             /*total number of independant samples*/


#define timePerCycle    (2.*PI/w)
#define timePerStep     (timePerCycle/TSTEPSPERC)

/* At time TTEST x must lie within these bounds.*/
#define TTEST       ((int)( (double)TSTEPSPERC*(7.3/timePerCycle)) )
#define XTESTMIN    (-0.478) 
#define XTESTMAX    (-0.442)

#define XSTART      (-0.9)      /*Fixed value of x at t=0 */

/* These variables specify the history of the system */
double  x[TBINS+1];             /* x at time t              */
double  gr[TBINS+1];            /* Random force at time t   */
double  pf[TBINS+1];            /* Periodic force at time t */

int     tried[TBINS+1]= {0};    /* keep track of trials and failurs*/
int     failed[TBINS+1]= {0};

/*******************************************************************/

main(  int argc , char *argv[])
{
    int i,t;
    int sample;
    

    /************************
      INITILISATION
    ************************/

    /****** Parse Command Line *********/
    if(argc!=2) die("%s samples",argv[0]);
    sscanf(argv[1],"%d",&SAMPLES);

    
    /****** Echo paremeters, and print axis *******/
    printf("Duffing Path Sampling\t D=%lg\t A=%lg\t w=%lg\t SAMPLES=%d\n",
                D,A,w,SAMPLES);
    
    /*output: time axis */
    printf("\t\t");
    for(t=0;t<TBINS;t+=SKIP) printf("%lf\t",(double)t*timePerStep);
    printf("\n");
    
    
    /********* Init periodic force *********/
    for(t=0;t<TBINS;t++)
        pf[t] = timePerStep*A* cos( w* t*timePerStep);
    
    
    /******** Generate an Initial Path ********/
    
    printf("Init Steps %d\n", initPath() );
    fflush(stdout);

    relaxPath(SHOOTS*64);
    

    
    /***********************************
        Main Simulation Loop
    ************************************/
    for(sample=0;sample<SAMPLES;sample++)
    {   
        relaxPath(SHOOTS);
        
        printf("%d\t\t",sample);
        for(t=0;t<TBINS;t+=SKIP) printf("%lf\t",x[t]);
        printf("\n");
        fflush(stdout);
    }
    
    /****** Print rejection analysis ****/
    printf("\n\n Tried\t\t");
    for(t=0;t<TBINS;t+=SKIP) printf("%d\t",tried[t]);

    printf("\nFailed\t\t");
    for(t=0;t<TBINS;t+=SKIP) printf("%d\t",failed[t]);
        
    printf("\n");

}


/******************************************************
    Propagate the path from t=start to t=end 
    NOTE: Changes in position occur from x[start+1] to x[end]
******************************************************/

#define K(x)    ((x) - ((x)*(x)*(x)))   /*Duffing part of the force */

void propagate(int start, int end)
{
    int t;
    for(t=start;t<end;t++)
    {
        x[t+1] = x[t] + timePerStep* K(x[t])  + gr[t]+pf[t];
    }
}


/******************************************************
  Generate an initial path by forcing the end point 
*******************************************************/
int initPath()
{
    int init=0;
    
    double diff;            
    double oldDiff;

    int     t;          /*Timestep at which changes are made */
    double oldgr;       /*Storage for old gr. Needed if shoot fails */

    x[0]=XSTART;
    
    for(t=0;t<TBINS;t++)
                gr[t] =sqrt(D*timePerStep)*genRanGaussian(); 
                
    propagate(0,TBINS);
    oldDiff=fabs(x[TTEST]-XTESTMIN);
    
    
    do
    {
        init++;
        t = (int) ceil(genRanf()*TTEST);  /* choose gr[] to change*/
        oldgr = gr[t];
        gr[t] =sqrt(D*timePerStep)*genRanGaussian(); /*Generate a gr with correct variance*/        propagate(0,TTEST);                     /*propagate new path to test time*/

        propagate(0,TTEST);                     /*propagate new path to test time*/

        diff=fabs(x[TTEST]-XTESTMIN);
        if( diff>oldDiff) /*TEST*/
        {
            /*FAILED*/
            gr[t]=oldgr;
        } else {
            oldDiff=diff;
        }
        
    }   while( (x[TTEST]<XTESTMIN) || (x[TTEST]>XTESTMAX) );
    
    propagate(0,TBINS);
    
    return init;
    
}



/******************************************************
  Generate a sequence of new paths consistant with
    boundary conditions.
*******************************************************/
void relaxPath(int relax)
{
    int loop;
        
    int     t;          /*Timestep at which changes are made */
    double oldgr;           /*Storage for old gr. Needed if shoot fails */

    for(loop=0;loop<relax;loop++)
    {
        t = (int) ceil(genRanf()*TTEST);  /* choose gr[] to change*/
        oldgr = gr[t];
        gr[t] =sqrt(D*timePerStep)*genRanGaussian(); /*Generate a gr with correct variance*/
        tried[t]++;
        
        
        propagate(0,TTEST);                     /*propagate new path to test time*/
    
        if( (x[TTEST]<XTESTMIN) || (x[TTEST]>XTESTMAX)) /*TEST*/
        {
            /*FAILED*/
            gr[t]=oldgr;
            failed[t]++;
        }
    }
    
    /* We have generated a statistically independent path from 0 to TTEST.
        Now propagate it forward in time. */
    for(t=TTEST+1;t<TBINS;t++) gr[t] = sqrt(D*timePerStep)*genRanGaussian();
    propagate(0,TBINS);
}
    

/*******************************************************************/

void die(char *comment, ...)
{   
    va_list argp;

    va_start(argp, comment);
        vfprintf(stderr, comment, argp);
    va_end(argp);
    fprintf(stderr, "\n");
    
    abort();
}
